{
    "AWSTemplateFormatVersion": "2010-09-09",
    "Description": "TestTemplate",
    "Parameters": {
        "Param1": {
            "Type": "String"
        },
        "Param2": {
            "Default": "default",
            "Type": "CommaDelimitedList"
        }
    },
    "Resources": {
      "Dummy": {
        "Type": "AWS::CloudFormation::WaitConditionHandle"
      }
    },
    "Outputs": {
      "DummyId": {
        "Value": "dummy-{{ context.environment.foo }}-{{ variables.Param1 }}-{{ variables.bar }}-1234"
      }
    }
}
